! Simulation of genotype data for Colony analysis of parentage & sibship, and then for migration estimation
! Input file : User defined
! Output file : User defined
! Call external binaries: Colony2p.exe, BA3.exe, MigEst.exe
MODULE Variables
  IMPLICIT NONE
  INTEGER, PARAMETER :: I4B = SELECTED_INT_KIND(9)    ! -2147483648 ~ 2147483647
  INTEGER, PARAMETER :: I2B = SELECTED_INT_KIND(4)    ! -32768 ~ 32767
  INTEGER, PARAMETER :: I1B = SELECTED_INT_KIND(2)    ! -128 ~ 127
  INTEGER, PARAMETER :: SP = KIND(1.0)                ! Single Precision
  INTEGER, PARAMETER :: DP = KIND(1.0D0)              ! Double Precision
  INTEGER(I4B) :: NumReplicate, NumPop, NumPrnt, NumLoci, &
    NumAllele, IDUM, SampleSize(3), AllPopSampled, ParallelRun, &
    RunColony, BayesAss, NumPopSmp, NumPop1, NumUnSmpPop, CI95, MigModel
  INTEGER(I1B), ALLOCATABLE :: PrntGtype(:, :, :, :, :) !allele,locus,indiv,sex,pop
  INTEGER(I1B), ALLOCATABLE :: OffGtype(:, :, :, :)  !allele,locus,indiv,sex,pop
  INTEGER(I4B), ALLOCATABLE :: Pedigree(:, :, :)
  REAL(DP) :: MigRate, MutRate, MigRate0, FST, FstEst, PriorWeight
  REAL(DP), ALLOCATABLE :: MetaFre(:, :), PopFre(:, :, :), &
    X_Est(:, :, :, :), m_Est(:, :, :, :), XXEst(:, :, :), mmEst(:, :), &
    mmEst0(:, :)
  CHARACTER(Len = 99) :: OutputFile
END MODULE Variables
!
Program SimuGenotype
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, NG
  EXTERNAL :: ReadParameter, Initialize, StartPop, Mutation, Reproduction, &
    Sampling, ComputeFre, ComputeFst, ComputePopFre, ColonyInput, Colony, &
    EstimateMigration, OUTPUT, BayesAssInput, RunBayesAss, DateTime
  CALL DateTime(1)
  CALL ReadParameter
  CALL Initialize
  NG = Max(Min(4 * NumPrnt, Nint(1.0 / (1 - MigRate0))), 8)
  DO I = 1, NumReplicate
    CALL StartPop
    DO J = 1, NG          !Simulate 2Ne generations
      CALL Mutation
      CALL Reproduction
      CALL Replace
    END DO
    CALL Reproduction
    IF(BayesAss == 1) CALL BayesAssInput(I)
    CALL Sampling
    CALL ComputeFre
    CALL ComputeFst
    CALL ComputePopFre
    IF(RunColony > 0) THEN
      CALL ColonyInput
      CALL Colony
    END IF
    CALL EstimateMigration(I)
  END DO
  IF(BayesAss == 1) CALL RunBayesAss
  CALL OUTPUT
  CALL DateTime(2)
END Program SimuGenotype
!
SUBROUTINE ReadParameter
  USE Variables
  IMPLICIT NONE
  Character(Len = 200) :: C200, InFile
  Logical :: Flag
  CALL Get_command_argument(1, InFile)
  IF(LEN_TRIM(InFile) == 0) THEN
    WRITE(*, '(A)') 'No input file!'
    STOP
  ELSE
    Inquire(File = TRIM(InFile), Exist = Flag)
    IF(.NOT. Flag) THEN
      WRITE(*, '(A)') 'Input file: ' // TRIM(InFile) // ' not exists!'
      STOP
    END IF
  END IF      
  OPEN(10, File = Trim(InFile))
  READ(10, *) NumReplicate
  IF(NumReplicate < 1) THEN
    WRITE(*, '(A)') '#replicates < 1!'
    STOP
  END IF   
  READ(10, *) NumPop
  IF(NumPop < 2) THEN
    WRITE(*, '(A)') '#populations < 2!'
    STOP
  END IF  
  READ(10, *) NumPopSmp
  IF(NumPopSmp < 2) THEN
    WRITE(*, '(A)') '#sampled populations < 2!'
    STOP
  END IF
  IF(NumPopSmp > NumPop) THEN
    WRITE(*, '(A)')'# sampled pop > # pop!'
    STOP
  ELSE
    AllPopSampled = Merge(1, 0, NumPopSmp == NumPop)
  END IF
  NumPop1 = NumPopSmp + 1 - AllPopSampled
  READ(10, *) NumUnSmpPop    !BayesAss ONLY
  IF(NumUnSmpPop < 0) THEN
    WRITE(*, '(A)') '#assumed unsampled populations < 0!'
    STOP
  END IF
  READ(10, *) NumPrnt
  IF(NumPrnt < 1) THEN
    WRITE(*, '(A)') '#parents < 1!'
    STOP
  END IF
  READ(10, *) NumLoci
  IF(NumLoci < 1) THEN
    WRITE(*, '(A)') '#Loci < 1!'
    STOP
  END IF
  READ(10, *) NumAllele
  IF(NumAllele < 2) THEN
    WRITE(*, '(A)') '#alleles per locus < 2!'
    STOP
  END IF
  READ(10, *) SampleSize(1 : 3)
  IF(ANY(SampleSize(1 : 2) > NumPrnt)) THEN
    WRITE(*, '(A)')'Parent sample size > population size!'
    STOP
  ELSE IF(ALL(SampleSize(1 : 2) < 1)) THEN
    WRITE(*, '(A)')'No parents sampled from a population!'
    STOP
  ELSE IF(SampleSize(3) < 1) THEN
    WRITE(*, '(A)')'No offspring sampled from a population!'
    STOP
  ELSE IF(SampleSize(3) > 2 * NumPrnt) THEN
    WRITE(*, '(A)')'Offspring sample size too big!'
    STOP
  END IF
  READ(10, *) MigModel, MigRate
  C200 = ''
  IF(MigModel /= 1 .AND. MigModel /= 2) THEN
    C200 = 'Migration model must be Island (1) or Stepping stone (2) model!'
  ELSE IF(MigModel == 1) THEN    !Island model
    IF(1._DP - MigRate * (NumPop - 1) < 0.D0) &
      C200 = 'Migration rate too high for the # populations'      
  ELSE IF(AllPopSampled /= 1) THEN
    C200 = 'All populations must be sampled for stepping stone model!'
  ELSE
    IF(NumPop < 4) THEN
      C200 = '#populations must > 3'
    ELSE IF(MigRate >= 0.5) THEN
      C200 = 'Migration rate too high'
    END IF
  END IF
  IF(Len_Trim(C200) > 0) THEN
    WRITE(*, '(A)')Trim(C200)
    STOP
  END IF
  READ(10, *) MutRate
  IF(MutRate < 0) THEN
    WRITE(*, '(A)')'Mutation rate too low!'
    STOP
  END IF
  IF(MutRate > 0.5) THEN
    WRITE(*, '(A)')'Mutation rate too high!'
    STOP
  END IF  
  READ(10, *) ParallelRun
  IF(ParallelRun < 0 .OR. ParallelRun > 1) THEN
    WRITE(*, '(A)')'Indicator for parallel run must be 1 (yes) or 0 (no)!'
    STOP
  END IF  
  READ(10, *) OutputFile
  OutputFile = AdjustL(OutputFile)
  READ(10, *) RunColony
  IF(RunColony < 0 .OR. RunColony > 2) THEN
    WRITE(*, '(A)')'Indicator variable RunColony = 0/1/2 ONLY!'
    STOP
  END IF
  READ(10, *) BayesAss
  IF(BayesAss < 0 .OR. BayesAss > 1) THEN
    WRITE(*, '(A)')'Indicator RunBayesAss = 0/1 ONLY!'
    STOP
  END IF
  IF(BayesAss == 1 .AND. Sum(SampleSize(1 : 3)) > 2 * NumPrnt) THEN
    WRITE(*, '(A)')'Offspring sample size for BayesAss too large!'
    STOP
  END IF
  READ(10, *) PriorWeight
  IF(PriorWeight <= 0) THEN
    WRITE(*, '(A)')'Prior weight too small!'
    STOP
  END IF
  READ(10, *) CI95
  IF(CI95 < 0 .OR. CI95 > 1) THEN
    WRITE(*, '(A)')'Indicator for 95% CI = 0/1 ONLY!'
    STOP
  END IF
  CLOSE(10)
END SUBROUTINE ReadParameter
!
SUBROUTINE Initialize
  USE Variables
  IMPLICIT NONE
  INTEGER :: I
  I = 1 + RunColony + BayesAss
  Allocate(PrntGtype(2, NumLoci, NumPrnt, 2, NumPop), &
    OffGtype(2, NumLoci, NumPrnt * 2, NumPop), &
    MetaFre(NumAllele, NumLoci), PopFre(NumAllele, NumLoci, NumPopSmp), &
    X_Est(2, 2, NumPopSmp, 1 + RunColony), m_Est(2, NumPopSmp, NumPop1, I), &
    Pedigree(3, NumPrnt * 2, NumPop), XXEst(2, 2, 1 + RunColony), &
    mmEst(2, I), mmEst0(2, I))
  X_Est(:, :, :, :) = 0._DP
  m_Est(:, :, :, :) = 0._DP
  XXEst(:, :, :) = 0._DP
  mmEst(:, :) = 0._DP
  mmEst0(:, :) = 0._DP
  FstEst = 0.D0
  IDUM = -12345
  MigRate0 = 1._DP - Merge(MigRate * (NumPop - 1), 2._DP * MigRate, MigModel == 1)
  OPEN(10, File = Trim(OutputFile))
  CLOSE(10, STATUS = 'DELETE')
END SUBROUTINE Initialize
!
SUBROUTINE StartPop
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, M
  REAL(SP) :: RAN1
  EXTERNAL :: RAN1
  DO I = 1, NumPop
    DO J = 1, 2
      DO K = 1, NumPrnt
        DO L = 1, NumLoci
          DO M = 1, 2
            PrntGtype(M, L, K, J, I) = Int(Floor(RAN1() * NumAllele + 1.0), I1B)
          END DO
        END DO
      END DO
    END DO
  END DO
END SUBROUTINE StartPop
!
SUBROUTINE Mutation
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, M, N, II
  REAL :: poidev, xm
  REAL(SP) :: RAN1
  EXTERNAL :: poidev, RAN1
  xm = 2.0 * NumPrnt * REAL(MutRate)
  DO L = 1, NumLoci
    DO I = 1, NumPop
      DO J = 1, 2
        M = NINT(PoiDev(xm))
        DO II = 1, M
          K = Floor(RAN1() * NumPrnt + 1.0)
          N = Floor(RAN1() * 2.0 + 1.0)
          PrntGtype(N, L, K, J, I) = Int(Floor(RAN1() * NumAllele + 1.0), I1B)
        END DO
      END DO
    END DO
  END DO
END SUBROUTINE Mutation
!
SUBROUTINE Reproduction
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, K, L, M, N, IP(2)
  REAL(SP) :: RAN1, R, CP(NumPop)
  EXTERNAL :: RAN1  
  DO I = 1, NumPop
    IF(MigModel == 1) THEN
      CP(:) = MigRate
    ELSE
      CP(:) = 0._DP
      IF(I == 1) THEN
        IP(:) = (/2, NumPop/)
      ELSE IF(I == NumPop) THEN
        IP(:) = (/1, NumPop - 1/)
      ELSE
        IP(:) = (/I - 1, I + 1/)
      END IF
      CP(IP(:)) = MigRate
    END IF
    CP(I) = MigRate0
    DO K = 2, NumPop - 1
      CP(K) = CP(K - 1) + CP(K)
    END DO
    CP(NumPop) = 1.1
    DO K = 1, NumPrnt * 2
      R = RAN1()
      DO N = 1, NumPop
        IF(R > CP(N)) CYCLE
        DO L = 1, 2
          IP(L) = Floor(RAN1() * NumPrnt + 1.0)
          Pedigree(L, K, I) = IP(L)
        END DO
        Pedigree(3, K, I) = N
        DO L = 1, NumLoci
          DO M = 1, 2
            OffGtype(M, L, K, I) = &
              PrntGtype(NINT(RAN1()) + 1, L, IP(M), M, N)
          END DO
        END DO
        EXIT
      END DO
    END DO
  END DO
END SUBROUTINE Reproduction
!
SUBROUTINE Replace
  USE Variables
  IMPLICIT NONE
  PrntGtype(:, :, :, 1, :) = OffGtype(:, :, 1 : NumPrnt, :)
  PrntGtype(:, :, :, 2, :) = OffGtype(:, :, 1 + NumPrnt :, :)
END SUBROUTINE Replace
!
SUBROUTINE Sampling
  USE Variables
  IMPLICIT NONE
  REAL(SP) :: RAN1
  INTEGER(I4B) :: I, J, K, L
  EXTERNAL :: RAN1  
  DO I = 1, NumPop
    DO J = 1, 2
      DO K = 1, NumPrnt - SampleSize(J)
        DO
          L = Floor(RAN1() * NumPrnt + 1.0)
          IF(PrntGtype(1, 1, L, J, I) < 1) CYCLE
          PrntGtype(1, 1, L, J, I) = 0
          EXIT
        END DO
      END DO
    END DO
    DO K = 1, 2 * NumPrnt - SampleSize(3)
      DO
        L = Floor(RAN1() * NumPrnt * 2 + 1.0)
        IF(OffGtype(1, 1, L, I) < 1) CYCLE
        OffGtype(1, 1, L, I) = 0
        EXIT
      END DO
    END DO
  END DO
END SUBROUTINE Sampling
!
Subroutine BayesAssInput(IR)
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, M, IR
  REAL(SP) :: RAN1
  CHARACTER(Len = 8) :: C8
  EXTERNAL :: RAN1
  WRITE(C8, '(I6)') IR
  OPEN(10, File = 'BA3in_' // Trim(AdjustL(C8)) // '.txt')
  M = 0
  DO I = 1, NumPopSmp
    DO K = 1, Sum(SampleSize(1 : 3))
      M = M + 1
      DO
        L = Floor(RAN1() * NumPrnt * 2 + 1.0)
        IF(OffGtype(1, 1, L, I) < 1_I1B) CYCLE
        OffGtype(1, 1, L, I) = -OffGtype(1, 1, L, I)
        EXIT
      END DO
      DO J = 1, NumLoci
        WRITE(10, '(5I6)') M, I, J, Abs(OffGtype(1 : 2, J, L, I))
      END DO
    END DO
  END DO
  IF(AllPopSampled == 0) THEN
    DO L = 1, NumUnSmpPop
      DO K = M + 1, M + Sum(SampleSize(1 : 3))
        DO J = 1, NumLoci
          WRITE(10, '(5I6)') K, NumPopSmp + L, J, 0, 0
        END DO
        M = M + 1
      END DO
    END DO
  END IF
  CLOSE(10)
  OffGtype(1, 1, :, :) = ABS(OffGtype(1, 1, :, :))
END Subroutine BayesAssInput
!
SUBROUTINE ComputeFre
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, M, Idx(NumAllele)
  REAL(DP) :: T
  DO L = 1, NumLoci
    MetaFre(:, L) = 0._DP
    DO I = 1, NumPopSmp
      Idx(:) = 0
      DO J = 1, 2 * NumPrnt
        IF(OffGtype(1, 1, J, I) < 1) CYCLE
        DO K = 1, 2
          Idx(OffGtype(K, L, J, I)) = Idx(OffGtype(K, L, J, I)) + 1
        END DO
      END DO
      DO J = 1, 2
        IF(SampleSize(J) < 1) CYCLE
        DO M = 1, NumPrnt
          IF(PrntGtype(1, 1, M, J, I) < 1) CYCLE
          DO K = 1, 2
            Idx(PrntGtype(K, L, M, J, I)) = Idx(PrntGtype(K, L, M, J, I)) + 1
          END DO
        END DO
      END DO
      T = 1._DP / Sum(Idx(:))
      PopFre(:, L, I) = Idx(:) * T
      MetaFre(:, L) = MetaFre(:, L) + PopFre(:, L, I) / NumPopSmp
    END DO
  END DO
END SUBROUTINE ComputeFre
!
SUBROUTINE ComputeFst
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, L, SmplSize
  REAL(DP) :: HS, HT, P(NumAllele), H(2)
! Gst from marker
  SmplSize = Sum(SampleSize(:)) * 2
  H(:) = 0.D0
  DO L = 1, NumLoci
    HS = NumPopSmp
    P(:) = 0._DP
    DO I = 1, NumPopSmp
      DO J = 1, NumAllele
        P(J) = P(J) + PopFre(J, L, I)
        HS = HS - PopFre(J, L, I) ** 2
      END DO
    END DO
    HS = HS / NumPopSmp
    HS = HS * SmplSize / (SmplSize - 1.D0)
    HT = 1.D0 - Sum((P(1 : NumAllele) / NumPopSmp) ** 2)
    HT = HT + HS / (SmplSize * NumPopSmp)
    H(1 : 2) = H(1 : 2) + (/HS, HT/)
  END DO
  Fst = (H(2) - H(1)) / H(2)
  IF(Fst < 0.00001D0) Fst = 0.00001D0
  FstEst = FstEst + FST
END SUBROUTINE ComputeFst
!
SUBROUTINE ComputePopFre
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, L, N
  REAL(DP) :: G
  N = Sum(SampleSize(1 : 3)) * 2
  G = (1.D0 - Fst) / Fst
  DO I = 1, NumPopSmp  
    DO L = 1, NumLoci
      DO J = 1, NumAllele
        PopFre(J, L, I) = (N * PopFre(J, L, I) + G * MetaFre(J, L)) / (N + G)
      END DO
    END DO
  END DO
END SUBROUTINE ComputePopFre
!
SUBROUTINE ColonyInput
  USE Variables
  IMPLICIT NONE
  CHARACTER(LEN = 11), PARAMETER :: FM1 = '(I10,T30,A)', FM2 = '(/I6,T30,A)'
  INTEGER(I4B) :: I, J, K, L, M, IP
  CHARACTER(LEN = 10) :: C10, D10, PopID
  CHARACTER(LEN = 99) :: C99
  DO IP = 1, NumPop
    WRITE(C10, '(I6)') IP
    OPEN(10, File='COLONY2_' // Trim(AdjustL(C10)) // '.DAT')
    WRITE(10, '(A,T30,A)')'Pop' // AdjustL(C10), '! C, Dataset name, Length<51'
    WRITE(10, '(A,T30,A)')'Pop' // AdjustL(C10), '! C, Main output file name, Length<21'
    WRITE(10, FM1)SampleSize(3), '! I, Number of offspring in the sample'
    WRITE(10, FM1)NumLoci, '! I, Number of loci'
    WRITE(10, FM1)12344 + RunColony, '! I, Seed for random number generator'
    WRITE(10, FM1)0, '! B, 0/1=Not updating/updating allele frequency'
    WRITE(10, FM1)2, '! I, 1/2=Monoecious/Dioecious'
    WRITE(10, FM1)0, '! B, 1/0=Y/N allow inbreeding'
    WRITE(10, FM1)0,'! B, 0/1=Diploid species/HaploDiploid species'
    WRITE(10, '(2I5,T30,A)')0, 0, '! B, 0/1=Polygamy/Monogamy for males & females'
    WRITE(10, '(I3,2F8.2,T30,2A)')2, 0., 0., '! I,R,R : 0/1/2=No/Ewen/Complexity prior', &
      ' =1, give mean paternal, maternal sibship size'    
    WRITE(10, FM1)1, '! B, 0/1=Unknown/Known population allele frequency'
    WRITE(10, '(999I4)')(NumAllele, I = 1, NumLoci)
    DO I = 1, NumLoci
      WRITE(10,'(999I8)')(J, J = 1, NumAllele)
      WRITE(10,'(999F8.4)')(PopFre(J, I, IP), J = 1, NumAllele)
    END DO
    WRITE(10, FM1)1, '! I, Number of runs'
    WRITE(10, FM1) 1, '! I, Length of Run (1, 2, 3) = (Short, Medium, Long)'
    WRITE(10, FM1) 0, '! B, 0/1=Monitor method by Iterate#/Time in second'  
    WRITE(10, FM1) 100000, '! I, Monitor interval in Iterate#/Seconds'
    WRITE(10, FM1) 0, '! B, 0/1=Other platform/Windows execution'
    WRITE(10, FM1) 1, '! B, 0/1/2=PairLikelihood score/Fulllikelihood/Combined Method'
    WRITE(10, FM1) 2, '! I, 1/2/3=Low/Medium/high precision with Fulllikelihood'
    WRITE(10, *)
    DO I = 1, NumLoci
      WRITE(C10, '(I6)') I
      WRITE(10, '(A)', ADVANCE = 'NO') ' Marker' // Trim(AdjustL(C10)) 
    END DO
    WRITE(10, *)
    WRITE(10, '(999I3)')(0, J = 1, NumLoci)
    DO M = 1, 2
      WRITE(10, '(999F4.0)')(0., J = 1, NumLoci)
    END DO
    K = 0
    DO I = 1, 2 * NumPrnt
      IF(OffGtype(1, 1, I, IP) < 1_I1B) CYCLE
      K = K + 1
      WRITE(C99, '(I6)') K
      WRITE(D10, '(I6)') Pedigree(3, I, IP)
      C99 = 'O' // Trim(AdjustL(C99)) // '.' // Trim(AdjustL(D10))
      DO M = 1, 2
        WRITE(D10, '(I6)') Pedigree(M, I, IP)
        C99 = Trim(AdjustL(C99)) // '.' // Trim(AdjustL(D10))
      END DO
      WRITE(10, '(A,999I4)')Trim(AdjustL(C99)) // ',', (OffGtype(1 : 2, L, I, IP), L = 1, NumLoci)
    END DO
    WRITE(10, *)
    WRITE(10, '(2F9.4,T30,A)') SampleSize(1 : 2) / DBLE(NumPrnt), &
      '!Prob that the dad & mum of an offspring included in candidates'
    WRITE(10, '(2I9,T30,A)') SampleSize(1 : 2) * NumPop, &
    '  !Numbers of candidate males and females'
    DO K = 1, 2
      WRITE(10,*)
      DO I = 1, NumPop
        WRITE(PopID, '(I6)') I
        PopID = Trim(AdjustL(PopID)) // '.'
        DO J = 1, NumPrnt
          IF(PrntGtype(1, 1, J, K, I) < 1_I1B) CYCLE
          WRITE(C10, '(I6)') J
          C10 = Trim(AdjustL(PopID)) // Trim(AdjustL(C10)) // '.' // Merge('m', 'f', K == 1)
          WRITE(10, '(A,999I4)') Trim(AdjustL(C10)) // ',', (PrntGtype(1 : 2, L, J, K, I), L = 1, NumLoci)
        END DO
      END DO
    END DO
    WRITE(10,*)
    DO J = 1, 2
      WRITE(10, FM2)0, '!Number of offspring with known ' // Merge('fa', 'mo', J == 1) // 'ther'
    END DO
    WRITE(10,*)
    DO J = 1, 2
      WRITE(10,FM2) 0, '!Number of known ' // Merge('p', 'm', J == 1) // 'aternal sibships'
    END DO
    DO I = 1, 2   !Output excluded candidates
      WRITE(10,FM2) 0, '!Number of offspring with known excluded candidates as ' &
        // Merge('fa', 'mo', I == 1) // 'ther'
    END DO
    WRITE(10,FM2) 0, '!Number of offspring with known excluded paternal sibships'
    WRITE(10,FM2) 0, '!Number of offspring with known excluded maternal sibships'
    CLOSE(10)
  END DO
END SUBROUTINE ColonyInput
!
Subroutine Colony
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: IP, NumThread
  CHARACTER(LEN = 19) :: C19
  CHARACTER(LEN = 2) :: C2
! OpenMP parallel for multiple processors
  NumThread = 1
!$ IF(ParallelRun > 0) NumThread = NumPopSmp
!$ Call OMP_SET_NUM_THREADS(NumThread)
!$OMP PARALLEL DEFAULT(SHARED) private(IP, C19, C2)
!$OMP DO SCHEDULE(dynamic, 1) 
  DO IP = 1, NumPopSmp
    WRITE(C19, '(I6)') IP
    CALL SYSTEM('Del Pop' // Trim(AdjustL(C19)) // '.*')
    C2 = Merge(' 1', ' 0', NumThread > 1 .AND. IP /= 1)
    C19 = 'COLONY2_' // Trim(AdjustL(C19)) // '.DAT'
    CALL SYSTEM('Colony2p ' // Trim(AdjustL(C19)) // C2)
!$  IF(NumThread > 1 .AND. C2 == ' 0') WRITE(*,'(A)') 'Waiting for other threads to complete'
  END DO
!$OMP END DO
!$OMP END PARALLEL
END Subroutine Colony
!
Subroutine RunBayesAss
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, NumThread
  CHARACTER(LEN = 19) :: C19
  CHARACTER(LEN = 2) :: C2
  CHARACTER(Len = 880) :: C880
  REAL(DP) :: BA3(NumPopSmp, NumPop1), T, Y
  EXTERNAL :: mmMSE
! OpenMP parallel for multiple processors
  NumThread = 1
!$ IF(ParallelRun > 0) NumThread = Min(NumPopSmp, NumReplicate)
!$ Call OMP_SET_NUM_THREADS(NumThread)
!$OMP PARALLEL DEFAULT(SHARED) private(I, C19, C2)
!$OMP DO SCHEDULE(dynamic, 1)
  DO I = 1, NumReplicate
    WRITE(C19, '(I6)') I
    C19 = '_' // Trim(AdjustL(C19)) // '.txt'
    C2 = Merge('  ', '-v', NumThread > 1 .AND. Mod(I, NumThread) /= 0)
    CALL SYSTEM('BA3 ' // C2 // ' -o BA3out' // Trim(C19) // ' BA3in' // Trim(C19))
  END DO
!$OMP END DO
!$OMP END PARALLEL
  OPEN(7, File = Trim(OutputFile), POSITION = 'APPEND')
  WRITE(7, '(//A, I6)') 'Estimate of BayesAss, Offspring sample size=', Sum(SampleSize(1:3))
  DO I = 1, NumReplicate
    WRITE(C19, '(I6)') I
    C19 = AdjustL(C19)
    OPEN(10, file ='BA3out_' // Trim(C19) // '.txt')
    DO
      READ(10, *) C880
      C880 = AdjustL(C880)
      IF(C880(1 : 9) == 'Migration') EXIT
    END DO
    READ(10, '(A)') C880
    DO L = 1, NumPopSmp
      READ(10, '(A)') C880
      K = 0
      DO J = 1, NumPopSmp
        K = INDEX(C880, ':')
        READ(C880(2 + K : 7 + K), *) BA3(L, J)
        C880 = C880(K + 8 :)
      END DO
      IF(NumPopSmp < NumPop) BA3(L, NumPop1) = 1 - Sum(BA3(L, 1 : NumPopSmp))
    END DO
    CLOSE(10)
    L = UBound(m_Est, 4)
    CALL mmMSE(BA3, L)
    WRITE(7, '(99F9.4)')
    DO J = 1, NumPopSmp
      WRITE(7, '(99F9.4)') BA3(J, :)
    END DO
  END DO
  CLOSE(7)
END Subroutine RunBayesAss
!
Subroutine EstimateMigration(IR)
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, M, IP, IR, Idx(NumPop, 3), Jdx(3)
  INTEGER(I1B) :: Kdx(2, NumPrnt * 2)
  REAL(DP) :: T, Y, XX_Est(2, NumPopSmp), mm_Est(NumPopSmp, NumPop1), &
    X_CI95(2, 2, NumPopSmp), m_CI95(2, NumPopSmp, NumPopSmp)
  CHARACTER(LEN = 99) :: C99, D99, E99, ID(2), Method
  LOGICAL :: ExistQ
  EXTERNAL :: mmMSE
  WRITE(*, '(/A,I6)') 'Replicate=', IR
  OPEN(7, File = Trim(OutputFile), POSITION = 'APPEND')
  WRITE(7, '(A,I4)')'Input & output from replicate', IR
  DO L = 1, 1 + RunColony
    OPEN(11, File = 'Parameter.txt')
    WRITE(11, '(I6,T30,A)') NumPopSmp, '!Number of sampled populations'
    WRITE(11, '(I6,T30,A)') AllPopSampled, '!1/0=Y/N: All populations sampled'
    WRITE(11, '(F12.4,T30,A)') PriorWeight, '!Beta distribution weight'
    IF(L == 1) THEN
      Method = 'True parentage'
    ELSE
      Method = Merge('PL', 'FL', L == 2) // ' parentage'
    END IF
    Method = AdjustL(Method)
    WRITE(7, '(/A)')'Input and result from ' // Trim(Method)
    DO IP = 1, NumPopSmp
      Idx(:, :) = 0
      IF(L == 1) THEN         !True parentage
        DO I = 1, 2 * NumPrnt
          IF(OffGtype(1, 1, I, IP) < 1_I1B) CYCLE
          Jdx(1 : 3) = Pedigree(1 : 3, I, IP)
          IF(PrntGtype(1, 1, Jdx(1), 1, Jdx(3)) < 1_I1B) THEN
            IF(PrntGtype(1, 1, Jdx(2), 2, Jdx(3)) < 1_I1B) CYCLE
            K = 2
          ELSE
            K = Merge(1, 3, PrntGtype(1, 1, Jdx(2), 2, Jdx(3)) < 1_I1B)
          END IF
          Idx(Jdx(3), K) = Idx(Jdx(3), K) + 1
        END DO
      ELSE IF(L == 3) THEN    !FL parentage
        WRITE(C99, '(I6)') IP
        C99 = 'Pop' // Trim(AdjustL(C99)) // '.BestConfig'
        OPEN(10, File = Trim(AdjustL(C99)))
        READ(10, *) D99
        DO I = 1, SampleSize(3)
          READ(10, *) D99, ID(1 : 2)
          DO J = 1, 2
            ID(J) = Trim(AdjustL(ID(J)))
            Jdx(J) = Scan(ID(J), '.')
          END DO
          DO J = 1, 2
            IF(Jdx(J) == 0) CYCLE
            K = Jdx(J) - 1
            READ(ID(J)(1 : K), *) Jdx(J)
          END DO
          IF(ALL(Jdx(1 : 2) > 0) .AND. Jdx(1) == Jdx(2)) THEN
            Idx(Jdx(1), 3) = Idx(Jdx(1), 3) + 1
          ELSE
            DO K = 1, 2
              IF(Jdx(K) > 0) Idx(Jdx(K), K) = Idx(Jdx(K), K) + 1
            END DO
          END IF
        END DO
        CLOSE(10)
      ELSE                  !PL parentage
        Kdx(:, :) = 0_I1B
        DO I = 1, 2
          IF(SampleSize(I) < 1) CYCLE
          WRITE(C99, '(I6)') IP            
          C99 = 'Pop' // Trim(AdjustL(C99)) // '.Pairwise' // &
            Merge('P', 'M', I == 1) // 'aternity'
          OPEN(10, File = Trim(AdjustL(C99)))
          READ(10, *, IOSTAT = K) C99
          DO
            READ(10, *, IOSTAT = K) C99, D99, E99
            IF(K /= 0) EXIT
            E99 = AdjustL(E99)
            READ(E99(1 : Len_Trim(E99) - 1), *) M
            IF(M < 80) CYCLE
            C99 = Trim(AdjustL(C99))
            J = Scan(C99, '.')
            READ(C99(2 : J - 1), *) J
            D99 = Trim(AdjustL(D99))
            K = Scan(D99, '.')
            READ(D99(1 : K - 1), *) K
            Kdx(I, J) = Int(K, I1B)
          END DO
        END DO
        DO I = 1, NumPrnt * 2
          IF(ALL(Kdx(1 : 2, I) == 0_I1B)) CYCLE
          IF(ALL(Kdx(1 : 2, I) > 0_I1B)) THEN
            IF(Kdx(1, I) == Kdx(2, I)) THEN
              Idx(Kdx(1, I), 3) = Idx(Kdx(1, I), 3) + 1
            ELSE
              DO J = 1, 2
                Idx(Kdx(J, I), J) = Idx(Kdx(J, I), J) + 1
              END DO
            END IF
          ELSE IF(Kdx(1, I) > 0_I1B) THEN
            Idx(Kdx(1, I), 1) = Idx(Kdx(1, I), 1) + 1
          ELSE
            Idx(Kdx(2, I), 2) = Idx(Kdx(2, I), 2) + 1
          END IF
        END DO
      END IF
      WRITE(C99, '(I6)') IP
      C99 = AdjustL(C99)
      DO M = 7, 11, 4
        WRITE(M, '(/2F8.4,T30,A)') SampleSize(1 : 2) / DBLE(NumPrnt), &
          '!Proportion of male & female parents sampled from pop ' // C99
        WRITE(M, '(2I8,T30,A)') SampleSize(1 : 2), &
          '!# male & female parents sampled from pop' // C99
        DO K = 1, 3
          WRITE(M, '(99I4)') Idx(1 : NumPopSmp, K)
        END DO
        WRITE(M, '(I4,T30,A)') SampleSize(3), '!Total# offspring from pop ' // C99
      END DO
    END DO
    WRITE(11, '(I2,T30,A)') CI95, '!1/0=Y/N for CI95'
    CLOSE(11)
    OPEN(10, File = 'Migrate.Finish')
    CLOSE(10, Status = 'DELETE')   
    CALL SYSTEM('MigEst.exe')
    DO
      Inquire(FILE = 'Migrate.Finish', Exist = ExistQ)
      IF(ExistQ) EXIT
    END DO
    OPEN(10, File = 'Migrate.out')
    READ(10, *) C99
    READ(10, *) C99
    DO I = 1, NumPopSmp
      READ(10, *) XX_Est(1 : 2, I)
    END DO
    READ(10, *) C99
    DO I = 1, NumPopSmp
      READ(10, *) mm_Est(I, :)
    END DO
    IF(CI95 == 1)THEN
      DO J = 1, 2
        READ(10, *) C99
        READ(10, *) C99
        DO I = 1, NumPopSmp
          IF(J == 1) THEN 
            READ(10, *) (X_CI95(1 : 2, K, I), K = 1, 2)
          ELSE
            READ(10, *) (m_CI95(1 : 2, I, K), K = 1, NumPopSmp)
          END IF
        END DO
      END DO
    END IF
    CLOSE(10)
    WRITE(7, '(/A,I4)')'Estimates from ' // Trim(Method)
    WRITE(7, '(A)')'Sampling proportions'
    DO J = 1, NumPopSmp
      WRITE(7, '(99F8.4)')XX_Est(:, J)
      X_Est(1, :, J, L) = X_Est(1, :, J, L) + XX_Est(:, J)
      X_Est(2, :, J, L) = X_Est(2, :, J, L) + &
        (XX_Est(:, J) - SampleSize(1 : 2) / DBLE(NumPrnt)) ** 2
      DO K = 1, 2
        IF(SampleSize(K) < 1) CYCLE
        Y = DBLE(SampleSize(K)) / NumPrnt
        XXEst(1, K, L) = XXEst(1, K, L) + XX_Est(K, J)
        XXEst(2, K, L) = XXEst(2, K, L) + (XX_Est(K, J) - Y) ** 2
      END DO
    END DO
    IF(CI95 == 1)THEN
      WRITE(7, '(/A)')'Sampling proportions: CI95'
      WRITE(7, '(6A14)') 'Population', &
        'CI_L(Male)', 'CI_U(Male)', 'CI_L(Female)', 'CI_U(Female)'
      DO I = 1, NumPopSmp
        WRITE(7, '(I14, 99F14.4)') I, (X_CI95(:, K, I), K = 1, 2)
      END DO
    END IF
    CALL mmMSE(mm_Est, L)
    WRITE(7, '(A)')'Immigration rates'
    DO J = 1, NumPopSmp
      Where(mm_Est(J, :) < 0.00011) mm_Est(J, :) = 0.0
      WRITE(7, '(99F8.4)')mm_Est(J, :)
    END DO
    IF(CI95 == 1)THEN
      WRITE(7, '(A)')'Immigration rates: CI95'
      WRITE(7, '(2A14)', Advance = 'NO') 'Population'
      DO I = 1, NumPopSmp
        WRITE(C99, '(I6)')I
        C99 = '(:,' // Trim(AdjustL(C99)) // ')'
        WRITE(7, '(2A14)', Advance = 'NO')'CI_L' // Trim(C99), 'CI_U' // Trim(C99)
      END DO
      WRITE(7, *)
      DO I = 1, NumPopSmp
        Where(m_CI95(:, I, 1 : NumPopSmp) < 0.00011) m_CI95(:, I, 1 : NumPopSmp) = 0.0
        WRITE(7, '(I14, 99F14.4)') i, (m_CI95(:, I, K), K = 1, NumPopSmp)
      END DO
    END IF
    WRITE(7, *)
  END DO
  CLOSE(7)
END Subroutine EstimateMigration
!
Subroutine mmMSE(Est, L)
  USE VARIABLES
  IMPLICIT NONE
  REAL(DP) :: Est(NumPopSmp, NumPop1), T, Y
  INTEGER :: J, K, L
  DO J = 1, NumPopSmp
    m_Est(1, J, :, L) = m_Est(1, J, :, L) + EST(J, :)
    DO K = 1, NumPop1
      IF(MigModel == 1) THEN
        T = EST(J, K) - Merge(MigRate, MigRate0, K /= J)
      ELSE
        IF(K == J) THEN
          T = EST(J, K) - MigRate0
        ELSE IF(J == 1) THEN
          T = EST(J, K) - Merge(MigRate, 0._DP, K == 2 .OR. K == NumPopSmp)
        ELSE IF(J == NumPopSmp) THEN
          T = EST(J, K) - Merge(MigRate, 0._DP, K == 1 .OR. K == J - 1)
        ELSE
          T = EST(J, K) - Merge(MigRate, 0._DP, K == J - 1 .OR. K == J + 1)
        END IF
      END IF
      m_Est(2, J, K, L) = m_Est(2, J, K, L) + T ** 2
      IF(J == K .OR. K > NumPopSmp) CYCLE
      mmEst(1, L) = mmEst(1, L) + EST(J, K)
      mmEst(2, L) = mmEst(2, L) + T ** 2       
    END DO
    IF(AllPopSampled == 1) CYCLE
    mmEst0(1, L) = mmEst0(1, L) + EST(J, NumPop1)
    mmEst0(2, L) = mmEst0(2, L) + &
     (EST(J, NumPop1) - MigRate * (NumPop - NumPopSmp)) ** 2
  END DO
END Subroutine mmMSE
!
Subroutine OUTPUT
  USE Variables
  IMPLICIT NONE
  INTEGER(I4B) :: I, J, K, L, M
  CHARACTER(LEN = 19) :: C19, D19, ID(NumPop1)
  DO I = 1, NumPop1
    WRITE(ID(I), '(I6)') I
  END DO
  FORALL(I = 1 : NumPop1) ID(I) = 'FromPop' // AdjustL(ID(I))
  OPEN(7, File = Trim(OutputFile), POSITION = 'Append')
  WRITE(7, '(/A)')'____________________________________________________________'
  FORALL(K = 1 : 1 + RunColony, I = 1 : NumPopSmp, J = 1 : 2)
    X_Est(:, J, I, K) = X_Est(:, J, I, K) / NumReplicate
    X_Est(2, J, I, K) = SQRT(ABS(X_Est(2, J, I, K)))
  END FORALL
  FORALL(K = 1 : 1 + RunColony + BayesAss, &
    I = 1 : NumPopSmp, J = 1 : NumPop1)
    m_Est(:, I, J, K) = m_Est(:, I, J, K) / NumReplicate
    m_Est(2, I, J, K) = SQRT(ABS(m_Est(2, I, J, K)))
  END FORALL
  XXEst = XXEst / Dble(NumReplicate * NumPopSmp)
  mmEst = mmEst / Dble(NumReplicate * NumPopSmp * (NumPopSmp - 1))
  IF(AllPopSampled == 0) mmEst0 = mmEst0 / Dble(NumReplicate * NumPopSmp)
  DO M = 1, 2
    DO L = 1, 1 + RunColony + BayesAss
      DO I = 6, 7
        IF(L == 1) THEN
          D19 = 'true parentage'
        ELSE IF(L == 2 .AND. RunColony > 0) THEN
          D19 = 'PL parentage'
        ELSE IF(L == 3 .AND. RunColony == 2) THEN
          D19 = 'FL parentage'
        ELSE
          D19 = 'BayesAss'          
        END IF
        WRITE(I, '(/A)') 'Estimates from ' // Trim(D19)  
        IF(M == 1) THEN
          IF(BayesAss /= 1 .OR. L < 2 + RunColony) THEN
            DO K = 1, 2
              WRITE(I, '(A)')Merge('Mean', 'RMSE', K == 1) // &
                ' estimates of sampling proportions'
              WRITE(I, '(3A8)') 'Pop', 'Male', 'Female'
              DO J = 1, NumPopSmp
                WRITE(I, '(I8, 2F8.4)') J, X_Est(K, :, J, L)
              END DO
            END DO
          END IF
          DO K = 1, 2
            WRITE(I, '(A)')Merge('Mean', 'RMSE', K == 1) // ' estimates of migration rates'
            WRITE(I, '(99A12)') 'Pop', (Trim(ID(J)), J = 1, NumPop1)
            DO J = 1, NumPopSmp
              WRITE(I, '(I12, 99F12.4)') J, m_Est(K, J, :, L)
            END DO
          END DO
        ELSE
          WRITE(I, '(/8A12)')'Mean(x1)', 'RMSE(x1)', 'Mean(x2)', 'RMSE(x2)', &
            'Mean(m)', 'RMSE(m)', 'Mean(m0)', 'RMSE(m0)'
          IF(BayesAss /= 1 .OR. L < 2 + RunColony) THEN          
            WRITE(I, '(8F12.4)')(XXEst(1, K, L), Sqrt(XXEst(2, K, L)), K = 1, 2), &
              mmEst(1, L), Sqrt(mmEst(2, L)), mmEst0(1, L), Sqrt(mmEst0(2, L))
          ELSE
            WRITE(I, '(8F12.4)')0., 0., 0., 0., &
              mmEst(1, L), Sqrt(mmEst(2, L)), mmEst0(1, L), Sqrt(mmEst0(2, L))
          END IF
        END IF
      END DO
    END DO
  END DO
  WRITE(7, '(/2A15)')'FST_Estimated', 'FST_Predicted'
  WRITE(7, '(2F15.4)') FstEst / NumReplicate, 1.D0 / (1.D0 + &
    8.D0 * NumPrnt * NumPop / (NumPop - 1.D0) * (NumPop * MigRate + MutRate))
  CLOSE(7)
END Subroutine OUTPUT
!
Subroutine DateTime(I)
  IMPLICIT NONE
  CHARACTER(Len = 8), Save :: Date2(2)
  CHARACTER(Len = 10), Save :: Time2(2)
  INTEGER :: I, J
  CALL DATE_AND_TIME(Date2(I), Time2(I))
  IF(I == 1) RETURN
  DO J = 1, 2  
    WRITE(*,'(A, ES13.6)') Merge('Started', 'Current', J == 1) &
      // ':D=' // Date2(J)(7:8) // '/' // Date2(J)(5:6) // &
      '/' // Date2(J)(1:4) // ' T=' // Time2(J)(5:6) // &
      '/' // Time2(J)(3:4) // '/' // Time2(J)(1:2)
  END DO
END Subroutine DateTime
!
FUNCTION poidev(xm)
  IMPLICIT NONE
  REAL poidev,xm,PI
  PARAMETER (PI=3.141592654)
  REAL alxm,em,g,oldm,sq,t,y,ran1,z
  Double Precision gammln
  SAVE alxm,g,oldm,sq
  DATA oldm /-1./
  if (xm<12.)then
    if (ABS(xm-oldm)>1.E-6) then
      oldm=xm
      g=exp(-xm)
    endif
    em=-1
    t=1.
2   em=em+1.
    t=t*ran1()
    if (t>g) goto 2
  else
    if (ABS(xm-oldm)>1.E-6) then
      oldm=xm
      sq=sqrt(2.*xm)
      alxm=log(xm)
      g=xm*alxm-REAL(gammln(Dble(xm)+1.D0))
    endif
1   y=tan(PI*ran1())
    em=sq*y+xm
    if (em<0.) goto 1
    em=int(em)
    Z = em*alxm-REAL(gammln(Dble(em)+1.D0))-g
    IF(Z < -87.0) goto 1
    t=0.90*(1.0+y**2)*exp(Z)
    if (ran1()>t) goto 1
  endif
  poidev=em
END FUNCTION poidev
!
FUNCTION Ran1()
  Use Variables, Only : IDUM
  IMPLICIT NONE
  INTEGER, PARAMETER :: K4B=selected_int_kind(9)
  REAL :: Ran1
  INTEGER(K4B), PARAMETER :: IA=16807,IM=2147483647,IQ=127773,IR=2836
  REAL, SAVE :: am
  INTEGER(K4B), SAVE :: ix=-1,iy=-1,k
  if (idum <= 0 .or. iy < 0) then
    am=nearest(1.0,-1.0)/Real(IM)
    iy=ior(ieor(888889999,abs(idum)),1)
    ix=ieor(777755555,abs(idum))
    idum=abs(idum)+1
  end if
  ix=ieor(ix,ishft(ix,13))
  ix=ieor(ix,ishft(ix,-17))
  ix=ieor(ix,ishft(ix,5))
  k=iy/IQ
  iy=IA*(iy-k*IQ)-IR*k
  if (iy < 0) iy=iy+IM
  Ran1=am*ior(iand(IM,ieor(ix,iy)),1)
END FUNCTION Ran1
!
FUNCTION gammln(xx)
  DOUBLE PRECISION gammln,xx
  INTEGER j
  DOUBLE PRECISION ser,stp,tmp,x,y,cof(6)
  SAVE cof,stp
  DATA cof,stp/76.18009172947146d0,-86.50532032941677d0, &
    24.01409824083091d0,-1.231739572450155d0,.1208650973866179d-2, &
    -.5395239384953d-5,2.5066282746310005d0/
  x=xx
  y=x
  tmp=x+5.5d0
  tmp=(x+0.5d0)*log(tmp)-tmp
  ser=1.000000000190015d0
  do j=1,6
    y=y+1.d0
    ser=ser+cof(j)/y
  end do
  gammln=tmp+log(stp*ser/x)
END FUNCTION gammln
          